# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"

import functools
import logging
import multiprocessing

from PyMca5.PyMcaPhysics.xas.XASClass import XASClass, e2k
from est.core.process.process import _input_desc
from est.core.process.process import _output_desc
from est.core.process.process import Process
from est.core.process.pymca.exafs import process_spectr_exafs
from est.core.types import XASObject, Spectrum

_logger = logging.getLogger(__name__)


def process_spectr_k(
    spectrum,
    configuration,
    overwrite=True,
    callbacks=None,
    output=None,
    output_dict=None,
):
    """

    :param spectrum: spectrum to process
    :type: :class:`.Spectrum`
    :param configuration: configuration of the pymca normalization
    :type: dict
    :param overwrite: False if we want to return a new Spectrum instance
    :type: bool
    :param callback: callback to execute.
    :param output: list to store the result, needed for pool processing
    :type: multiprocessing.manager.list
    :param output_dict: key is input spectrum, value is index in the output
                        list.
    :type: dict
    :return: processed spectrum
    :rtype: tuple (configuration, spectrum)
    """
    _logger.debug(
        "start k weight definition on spectrum (%s, %s)" % (spectrum.x, spectrum.y)
    )
    assert spectrum is not None

    pymca_xas = XASClass()
    if spectrum.energy is None or spectrum.mu is None:
        raise ValueError(
            "Energy and or Mu is/are not specified, unable to " "compute exafs"
        )
    pymca_xas.setSpectrum(energy=spectrum.energy, mu=spectrum.mu)
    if configuration is not None:
        if "KWeight" in configuration and configuration["KWeight"] is not None:
            configuration["FT"]["KWeight"] = configuration["KWeight"]
            configuration["EXAFS"]["KWeight"] = configuration["KWeight"]
        pymca_xas.setConfiguration(configuration)

    if overwrite:
        spectrum_ = spectrum
    else:
        spectrum_ = Spectrum.from_dict(spectrum.to_dict())

    # we need to update EXAFSNormalized since we are overwriting it
    cf, exafs_res = process_spectr_exafs(
        spectrum=spectrum_, configuration=configuration
    )
    if exafs_res is None:
        err = "Failed to process exafs."
        if spectrum.x is not None or spectrum.y is not None:
            err = (
                err
                + "Spectrum (x, y) coords: "
                + ",".join((str(spectrum.x), str(spectrum.y)))
            )
        raise ValueError(err)

    # update EXAFSNormalized
    e0 = pymca_xas.calculateE0()

    kValues = e2k(spectrum_.energy - e0)
    exafs = exafs_res["EXAFSNormalized"]
    if "KWeight" in configuration and configuration["KWeight"] is not None:
        exafs *= pow(kValues, configuration["KWeight"])
    spectrum_["EXAFSNormalized"] = exafs
    configuration_ = pymca_xas.getConfiguration()

    if callbacks:
        for callback in callbacks:
            callback()

    if output is not None:
        assert output_dict is not None
        output[output_dict[spectrum]] = spectrum_
    return configuration_, spectrum_


def pymca_k_weight(xas_obj):
    """
    Set weight for exafs values

    :param xas_obj: object containing the configuration and spectra to process
    :type: Union[:class:`.XASObject`, dict]
    :return: spectra dict
    :rtype: :class:`.XASObject`
    """
    k_weight_obj = PyMca_k_weight()
    return k_weight_obj.process(xas_obj=xas_obj)


_USE_MULTIPROCESSING_POOL = False
# note: we cannot use multiprocessing pool with pypushflow for now.


class PyMca_k_weight(Process):

    inputs = [
        _input_desc(name="xas_obj", type=XASObject, handler="process", doc=""),
    ]

    outputs = [
        _output_desc(name="xas_obj", type=XASObject, doc=""),
    ]

    def __init__(self):
        Process.__init__(self, name="k weight")

    def set_properties(self, properties):
        if "_kWeightSetting" in properties:
            _properties = properties.copy()
            _properties["k_weight"] = _properties["_kWeightSetting"]
            del _properties["_kWeightSetting"]
            self.setConfiguration(_properties)

    def _k_weight(self):
        if "k_weight" in self._settings:
            return self._settings["k_weight"]
        return None

    def process(self, xas_obj):
        """

        :param xas_obj: object containing the configuration and spectra to process
        :type: Union[:class:`.XASObject`, dict]
        :return: spectra dict
        :rtype: :class:`.XASObject`
        """
        assert xas_obj is not None
        _xas_obj = self.getXasObject(xas_obj=xas_obj)
        if self._k_weight() is not None:
            _xas_obj.configuration["SET_KWEIGHT"] = self._k_weight()

        if "SET_KWEIGHT" not in _xas_obj.configuration:
            _logger.warning(
                "Missing configuration to know which value we should set "
                "to k weight, will be set to 0 by default"
            )
            _xas_obj.configuration["SET_KWEIGHT"] = 0

        for key in ("FT", "EXAFS", "Normalization"):
            if key not in _xas_obj.configuration:
                _xas_obj.configuration[key] = {}

        _xas_obj.configuration["KWeight"] = _xas_obj.configuration["SET_KWEIGHT"]
        _xas_obj.configuration["FT"]["KWeight"] = _xas_obj.configuration["SET_KWEIGHT"]
        _xas_obj.configuration["EXAFS"]["KWeight"] = _xas_obj.configuration[
            "SET_KWEIGHT"
        ]
        _xas_obj.configuration["Normalization"]["KWeight"] = _xas_obj.configuration[
            "SET_KWEIGHT"
        ]

        self._advancement.reset(max_=_xas_obj.n_spectrum)
        self._advancement.startProcess()
        self._pool_process(xas_obj=_xas_obj)
        self._advancement.endProcess()
        self.register_process(xas_obj=_xas_obj, data_keys=tuple())
        return _xas_obj

    def _pool_process(self, xas_obj):
        """process normalization from a pool"""
        assert isinstance(xas_obj, XASObject)
        if not _USE_MULTIPROCESSING_POOL:
            assert "KWeight" in xas_obj.configuration
            for spectrum in xas_obj.spectra.data.flat:
                process_spectr_k(
                    spectrum=spectrum,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=True,
                )
                assert "KWeight" in xas_obj.configuration
        else:
            with multiprocessing.Pool(5) as p:
                partial_ = functools.partial(
                    process_spectr_k,
                    configuration=xas_obj.configuration,
                    callbacks=self.callbacks,
                    overwrite=True,
                )
                p.map(partial_, xas_obj.spectra)

    def definition(self):
        return "Define k weight for xas treatment"

    def program_version(self):
        import PyMca5

        return PyMca5.version()

    def program_name(self):
        return "pymca_k_weight"

    __call__ = process


if __name__ == "__main__":
    import sys
    import yaml

    xas_object_yaml_file = sys.argv[1]
    _logger.debug("Load xas object from {}".format(xas_object_yaml_file))
    with open(xas_object_yaml_file, "r") as file:
        ddict = yaml.load(file)["input_data"]
        xas_object = XASObject.from_dict(ddict)
    print("******* do k-weight ********")
    res_xas_object = pymca_k_weight(xas_obj=xas_object)
    res_xas_object._create_saving_pt()

    # dump resulting object
    with open(xas_object_yaml_file, "w") as file:
        yaml.dump({"input_data": res_xas_object.to_dict()}, file)

    # dump resulting object into output_normalization file
    with open("output_k_weight.yaml", "w") as file:
        yaml.dump({"input_data": res_xas_object.to_dict()}, file)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"


from Orange.widgets.widget import OWWidget
from Orange.widgets.widget import Input, Output
from est.core.types import XASObject
from silx.gui import qt
import logging

_logger = logging.getLogger(__file__)


class SavingPointOW(OWWidget):
    """Widget to save the treatment done up to a specific point in the
    treatment flow.
    """

    name = "saving point"
    id = "orange.widgets.xas.utils.saving_point"
    description = (
        "Save treatment to a specific point and allow user to take"
        " back treatment from this point"
    )
    icon = "icons/saving.png"
    priority = 8
    category = "esrfWidgets"
    keywords = ["spectroscopy", "saving point", "save", "reprocess", "restart"]

    want_main_area = True
    resizing_enabled = False
    allows_cycle = False

    class Inputs:
        xas_obj = Input("xas_obj", XASObject, default=True)

    class Outputs:
        xas_obj = Output("xas_obj", XASObject)

    def __init__(self):
        super().__init__()

        self._last_xas_object = None

        self._buttons = qt.QDialogButtonBox(parent=self)
        self._loadSavingPt = qt.QPushButton("activate saving point", parent=self)
        self._loadSavingPt.setToolTip(
            "Load last data received and past it to " "next processing(s)"
        )
        self._buttons.addButton(self._loadSavingPt, qt.QDialogButtonBox.ActionRole)
        self.layout().addWidget(self._buttons)
        self._loadSavingPt.released.connect(self.reloadLastXasObj)

    @Inputs.xas_obj
    def _accept(self, xas_obj):
        self._last_xas_object = xas_obj.copy(create_h5_file=True)
        self.accept()
        self.Outputs.xas_obj.send(xas_obj)

    def reloadLastXasObj(self):
        """
        Load last saved xas_object and pass it to the next processing points
        """
        if self._last_xas_object is not None:
            self.accept()
            self.Outputs.xas_obj.send(self._last_xas_object.copy(create_h5_file=True))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tools to visualize spectra"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/07/2019"


from silx.gui import qt
from silx.gui.plot.items.roi import RectangleROI
from silx.gui.plot.tools.roi import RegionOfInterestManager
from silx.gui.plot.tools.roi import RegionOfInterestTableWidget

from est.gui.XasObjectViewer import MapViewer


class ROISelector(qt.QWidget):
    """
    Used to defined a roi on a spectra map
    """

    def __init__(self, parent=None):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        # TODO: if data starts to be quiet large then we can load only part of
        # if. There is some code to do this on tomwer.
        self._xas_object_map = MapViewer(keys=["mu"])
        self.layout().addWidget(self._xas_object_map)
        self._plot = self._xas_object_map.getPlot()

        # Create the object controlling the ROIs and set it up
        self._roiManager = RegionOfInterestManager(self._plot)
        self._roiManager.setColor("red")  # Set the color of ROI

        self._roiManager.sigRoiAdded.connect(self.updateAddedRegionOfInterest)

        # Add a rectangular region of interest
        self._roi = RectangleROI()
        self._roi.setGeometry(origin=(0, 0), size=(2, 2))
        self._roi.setName("ROI")
        self._roiManager.addRoi(self._roi)
        self.__roi_first_definition = True
        """flag used make sure the roi is clearly visible at least during the 
        first processing"""

        # Create the table widget displaying
        self._roiTable = RegionOfInterestTableWidget()
        self._roiTable.setRegionOfInterestManager(self._roiManager)

        # Add the region of interest table and the buttons to a dock widget
        widget = qt.QWidget()
        layout = qt.QVBoxLayout()
        widget.setLayout(layout)
        layout.addWidget(self._roiTable)
        layout.addWidget(self._xas_object_map.keySelectionDocker)

        dock = qt.QDockWidget("Image ROI")
        dock.setWidget(widget)
        self._plot.addDockWidget(qt.Qt.RightDockWidgetArea, dock)

        # expose API
        self.setColormap = self._plot.setDefaultColormap

    def setXasObject(self, xas_obj):
        self._xas_object_map.setXasObject(xas_obj=xas_obj)
        if self.__roi_first_definition is True:
            origin = (0, 0)
            size = (xas_obj.dim2, xas_obj.dim1)
            self.setROI(origin=origin, size=size)
            self._roi.setEditable(True)

    def getXasObject(self):
        return self._xas_object_map._xasObj

    def updateAddedRegionOfInterest(self, roi):
        """Called for each added region of interest: set the name"""
        if roi.getLabel() == "":
            roi.setName("ROI %d" % len(self._roiManager.getRois()))

    def setROI(self, origin, size):
        self._roi.setGeometry(origin=origin, size=size)

    def getROI(self):
        return self._roi


class ROISelectorDialog(qt.QDialog):
    """Dialog embedding the SoiSelector"""

    def __init__(self, parent=None):
        qt.QDialog.__init__(self, parent)
        self._mainWindow = ROISelector(parent=self)

        self.setWindowTitle("Roi selection")
        self.setWindowFlags(qt.Qt.Widget)
        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        _buttons = qt.QDialogButtonBox(parent=self)
        _buttons.setStandardButtons(types)

        self.setLayout(qt.QVBoxLayout())
        self.layout().addWidget(self._mainWindow)
        self.layout().addWidget(_buttons)

        _buttons.accepted.connect(self.accept)
        _buttons.rejected.connect(self.reject)

        # expose API
        self.setXasObject = self._mainWindow.setXasObject
        self.getXasObject = self._mainWindow.getXasObject
        self.getROI = self._mainWindow.getROI
        self.setROI = self._mainWindow.setROI
        self.setColormap = self._mainWindow.setColormap


if __name__ == "__main__":
    from est.core.types import XASObject

    import numpy
    from est.core.utils import spectra as spectra_utils
    from est.core.process.roi import ROIProcess, _ROI
    from silx.gui.colors import Colormap

    energy, spectra = spectra_utils.create_dataset(shape=(256, 32, 64))
    xas_obj = XASObject(spectra=spectra, energy=energy, configuration=None)

    app = qt.QApplication([])

    class ROISelectorDialogTest(ROISelectorDialog):
        """Infinite update the xas obj according to ROI"""

        def accept(self):
            roi_process = ROIProcess()
            roi_ = _ROI.from_silx_def(self.getROI())
            roi_process.set_properties({"roi": roi_.to_dict()})
            update_xas_object = roi_process.process(self.getXasObject())
            self.setXasObject(update_xas_object)
            self.show()

    roiSelector = ROISelectorDialogTest()
    # keep colormap
    roiSelector.setColormap(Colormap(name="temperature", vmin=0, vmax=10))
    roiSelector.setXasObject(xas_obj=xas_obj)
    roiSelector.show()

    app.exec_()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "08/08/2019"

import logging
import os
import shutil
import tempfile
import unittest
from silx.gui import qt
import time
import h5py
from orangecontrib.est.test.OrangeWorkflowTest import OrangeWorflowTest

try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
from est.gui.qtapplicationmanager import QApplicationManager


logging.disable(logging.INFO)

_logger = logging.getLogger(__file__)

app = QApplicationManager()


@unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
class TestSimpleLarchWorkflow(OrangeWorflowTest):
    """Test the following workflow:
    - XASInputOW
    - NormalizationOW

    - XASOutputOW
    """

    def setUp(self):
        OrangeWorflowTest.setUp(self)
        # set input file
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        self.xasInputWidget.setFileSelected(data_file)
        # set output file
        self.outputdir = tempfile.mkdtemp()
        self.output_file = os.path.join(self.outputdir, "output.h5")
        self.xasOutputWidget.setFileSelected(self.output_file)

    def tearDow(self):
        if os.path.isdir(self.outputdir):
            shutil.rmtree(self.outputdir)
        OrangeWorflowTest.tearDown(self)

    @classmethod
    def setUpClass(cls):
        OrangeWorflowTest.setUpClass()
        # create widgets
        xasInputNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.utils.xas_input.XASInputOW"
        )
        xasPreEdgeNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.larch.pre_edge.PreEdgeOW"
        )
        xasAutobkNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.larch.autobk.AutobkOW"
        )
        xasXFTFNode = cls.addWidget(cls, "orangecontrib.est.widgets.larch.xftf.XFTFOW")
        xasMBackNormNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.larch.mback_norm.Mback_normOW"
        )
        xasOutputNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.utils.xas_output.XASOutputOW"
        )

        cls.processOrangeEvents(cls)

        cls.link(cls, xasInputNode, "xas_obj", xasPreEdgeNode, "xas_obj")
        cls.link(cls, xasPreEdgeNode, "xas_obj", xasAutobkNode, "xas_obj")
        cls.link(cls, xasAutobkNode, "xas_obj", xasXFTFNode, "xas_obj")
        cls.link(cls, xasXFTFNode, "xas_obj", xasMBackNormNode, "xas_obj")
        cls.link(cls, xasMBackNormNode, "xas_obj", xasOutputNode, "xas_obj")
        cls.processOrangeEvents(cls)

        cls.xasInputWidget = cls.getWidgetForNode(cls, xasInputNode)
        cls.xasPreEdgeWidget = cls.getWidgetForNode(cls, xasPreEdgeNode)
        cls.xasAutobkWidget = cls.getWidgetForNode(cls, xasAutobkNode)
        cls.xasXFTFWidget = cls.getWidgetForNode(cls, xasXFTFNode)
        cls.xasMbackNormWidget = cls.getWidgetForNode(cls, xasMBackNormNode)
        cls.xasOutputWidget = cls.getWidgetForNode(cls, xasOutputNode)

    @classmethod
    def tearDownClass(cls):
        cls.xasInputWidget = None
        cls.xasPreEdgeWidget = None
        cls.xasAutobkWidget = None
        cls.xasXFTFWidget = None
        cls.xasMbackNormWidget = None
        cls.xasOutputWidget = None
        OrangeWorflowTest.tearDownClass()

    def test(self):
        """Make sure the workflow is valid and end on the data transfert"""
        # start processing
        self.xasInputWidget.restart()
        # wait until end
        app = qt.QApplication.instance()
        while app.hasPendingEvents():
            app.processEvents()
            # wait for calculation finished
            time.sleep(0.1)
            self.processOrangeEventsStack()

        self.assertEqual(self.xasPreEdgeWidget._window.getNCurves(), 4)
        # check plot have been updated
        self.assertEqual(self.xasAutobkWidget._window.getNCurves(), 3)
        self.assertEqual(self.xasXFTFWidget._window.getNCurves(), 3)
        self.assertEqual(self.xasMbackNormWidget._window.getNCurves(), 2)
        self.assertTrue(os.path.exists(self.output_file))

        # test outputfile
        with h5py.File(self.output_file, "r") as hdf5:
            self.assertTrue("scan1" in hdf5)
            scan_grp = hdf5["scan1"]
            self.assertTrue("absorbed_beam" in scan_grp)
            self.assertTrue("xas_process_1" in scan_grp)
            self.assertTrue("xas_process_2" in scan_grp)
            self.assertTrue("xas_process_3" in scan_grp)
            self.assertTrue("xas_process_4" in scan_grp)
            process_1_grp = scan_grp["xas_process_1"]
            self.assertTrue(process_1_grp["program"][()] == "larch_pre_edge")
            process_4_grp = scan_grp["xas_process_4"]
            self.assertTrue(process_4_grp["program"][()] == "larch_mback_norm")


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestSimpleLarchWorkflow,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"

import logging
import os
import shutil
import tempfile
import unittest
from silx.gui import qt
from orangecontrib.est.test.OrangeWorkflowTest import OrangeWorflowTest

try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
from est.gui.qtapplicationmanager import QApplicationManager


logging.disable(logging.INFO)

_logger = logging.getLogger(__file__)

app = QApplicationManager()


@unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
class TestSimplePyMcaWorkflow(OrangeWorflowTest):
    """Test the following workflow:
    - XASInputOW
    - NormalizationOW

    - XASOutputOW
    """

    def setUp(self):
        OrangeWorflowTest.setUp(self)
        # set input file
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        self.xasInputWidget.setFileSelected(data_file)
        # set output file
        self.outputdir = tempfile.mkdtemp()
        self.output_file = os.path.join(self.outputdir, "output.h5")
        self.xasOutputWidget.setFileSelected(self.output_file)

    def tearDow(self):
        if os.path.isdir(self.outputdir):
            shutil.rmtree(self.outputdir)
        OrangeWorflowTest.tearDown(self)

    @classmethod
    def setUpClass(cls):
        OrangeWorflowTest.setUpClass()
        # create widgets
        xasInputNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.utils.xas_input.XASInputOW"
        )
        xasNormalizationNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.pymca.normalization.NormalizationOW"
        )
        xasEXAFSNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.pymca.exafs.ExafsOW"
        )
        xasKWeightNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.pymca.k_weight.KWeightOW"
        )
        xasFTNode = cls.addWidget(cls, "orangecontrib.est.widgets.pymca.ft.FTOW")
        xasOutputNode = cls.addWidget(
            cls, "orangecontrib.est.widgets.pymca.utils.xas_output.XASOutputOW"
        )

        cls.processOrangeEvents(cls)

        cls.link(cls, xasInputNode, "xas_obj", xasNormalizationNode, "xas_obj")
        cls.link(cls, xasNormalizationNode, "xas_obj", xasEXAFSNode, "xas_obj")
        cls.link(cls, xasEXAFSNode, "xas_obj", xasKWeightNode, "xas_obj")
        cls.link(cls, xasKWeightNode, "xas_obj", xasFTNode, "xas_obj")
        cls.link(cls, xasFTNode, "xas_obj", xasOutputNode, "xas_obj")
        cls.processOrangeEvents(cls)

        cls.xasInputWidget = cls.getWidgetForNode(cls, xasInputNode)
        cls.xasNormalizationWidget = cls.getWidgetForNode(cls, xasNormalizationNode)
        cls.xasEXAFSWidget = cls.getWidgetForNode(cls, xasEXAFSNode)
        cls.xasKWeightWidget = cls.getWidgetForNode(cls, xasKWeightNode)
        cls.xasFTWidget = cls.getWidgetForNode(cls, xasFTNode)
        cls.xasOutputWidget = cls.getWidgetForNode(cls, xasOutputNode)

    @classmethod
    def tearDownClass(cls):
        cls.xasInputWidget = None
        cls.xasNormalizationWidget = None
        cls.xasEXAFSWidget = None
        cls.xasKWeightWidget = None
        cls.xasFTWidget = None
        cls.xasOutputWidget = None
        OrangeWorflowTest.tearDownClass()

    def test(self):
        """Make sure the workflow is valid and end on the data transfert"""
        # start processing
        self.xasInputWidget.restart()
        # wait until end
        app = qt.QApplication.instance()
        while app.hasPendingEvents():
            app.processEvents()
            self.processOrangeEventsStack()

        # check plot have been updated
        self.assertEqual(self.xasNormalizationWidget._window.getNCurves(), 4)
        self.assertEqual(self.xasEXAFSWidget._window.getNCurves(), 3)
        self.assertEqual(self.xasKWeightWidget._window.getNCurves(), 2)
        self.assertEqual(self.xasFTWidget._window.getNCurves(), 2)

        #  check if file as been saved
        self.assertTrue(os.path.exists(self.output_file))


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestSimplePyMcaWorkflow,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite



# -*- coding: utf-8 -*-
#
# This file is part of the bliss project
#
# Copyright (c) 2015-2020 Beamline Control Unit, ESRF
# Distributed under the GNU LGPLv3. See LICENSE for more info.

"""Physical units for engineering and science

Implementation based on :mod:`pint`.

Usage::

    >>> from bliss.physics.units import ur

    >>> # use ur as a pint UnitRegistry
    >>> mass = 0.1*ur.mg
    >>> E = mass * ur.c**2
    >>> print( E.to(ur.kJ) )
    8987551.78737 kilojoule

    >>> # decorate your methods to ensure proper units
    >>> @ur.units(mass='kg', result='J')
    ... def energy(mass):
    ...     return mass * ur.c**2

    >>> # passing a Quantity will give you back a Quantity
    >>> print( energy(mass) )
    8987551787.37 joule

    >>> # passing a float will assume it is in the proper (kg)
    >>> # and will give you back a float in the corresponding unit (J)
    >>> print( energy(0.1e-6) )
    8987551787.37


Advanced usage
--------------

Use another UnitRegistry
~~~~~~~~~~~~~~~~~~~~~~~~

Bliss unit system creates a default UnitRegistry (ur).

If you want to interact with another library which also uses pint it is
important that both libraries use the same UnitRegistry.

This library allows you to change the active UnitRegistry. You should do
this as soon as possible in the code of your application::

    import pint
    ureg = UnitRegistry()

    from bliss.physics import units
    units.ur = ureg

"""

from functools import wraps
from inspect import getfullargspec

import pint

__all__ = ["ur", "units"]

#: unit registry
ur = pint.UnitRegistry()


def is_quantity(arg):
    """Return whether the given argument is a quantity."""
    return isinstance(arg, ur.Quantity)


def to_unit(arg):
    """Permissively cast the given argument into a unit."""
    return ur.Unit(arg) if arg else None


def values_to_units(dct):
    """Cast the values of the given dict into units"""
    return {k: to_unit(v) for k, v in dct.items()}


def convert_to(arg, unit):
    """Permissively convert the given argument into the given unit.

    The argument can either be a number or a quantity.
    """
    if not unit:
        return arg
    if arg is None:
        return None
    else:
        return arg.to(unit) if is_quantity(arg) else arg * unit


def units(**kwarg_units):
    """
    Use as a decorator to protect your function against unit errors.

    Each keyword argument must be an argument of the function. The
    value is the unit (string or Unit) in which your function argument
    should be called with.

    An extra argument *result* should provide the Unit for the expected
    return value. And yes, you cannot have a function which has an
    argument called *result* but that is just good naming!

    Missing arguments will be ignored.

    Example::

        from bliss.physics.units import ur, units

        @units(mass='kg', result=ur.J)
        def energy(mass):
            return mass * ur.c**2

    When you call a decorated function, it will return a Quantity if the
    *result* is a Unit and at least one of the arguments is a Quantity. If none
    of the arguments is a Quantity the result is a float with a value in the
    units specified by *result*
    """
    result_unit = to_unit(kwarg_units.pop("result", None))
    kwarg_units = values_to_units(kwarg_units)

    def decorator(func):
        arg_spec = getfullargspec(func).args
        if not set(arg_spec).issuperset(kwarg_units):
            raise TypeError("units argument names differ from function argument names")

        @wraps(func)
        def wrapper(*args, **kwargs):
            # Everything is a kwargs
            kwargs.update(zip(arg_spec, args))
            # Check for quantity-free use case
            all_magnitude = all(
                not is_quantity(value)
                for key, value in kwargs.items()
                if key in kwarg_units
            )
            # Kwargs conversion
            kwargs = {
                key: convert_to(value, kwarg_units.get(key))
                for key, value in kwargs.items()
            }
            # Call the actual func
            result = func(**kwargs)
            if not result_unit:
                return result
            # Safety check
            if not is_quantity(result):
                raise TypeError("Function {!r} did not return a quantity".format(func))
            # Convert the result and return magnitude or quantity
            result = convert_to(result, result_unit)
            return result.magnitude if all_magnitude else result

        return wrapper

    return decorator



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/26/2019"


import numpy
import os
import unittest
from est.core.types import Spectrum
from est.units import ur
from est.core.types import Dim

try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
    from est.io.utils.spec import read_spectrum


class TestSpectrum(unittest.TestCase):
    """Test the spectrum class"""

    @unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
    def test_from_dat(self):
        """check that we can create a Spectrum from a pymca .dat file"""
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        energy, mu = read_spectrum(
            data_file,
            energy_col_name="Column 1",
            absorption_col_name="Column 2",
        )
        spectrum = Spectrum(energy=energy, mu=mu)
        self.assertTrue(spectrum.energy is not None)
        self.assertTrue(spectrum.mu is not None)

    def test_from_numpy_array(self):
        """check that we can create a Spectrum from numpy arrays"""
        energy = numpy.arange(10, 20)
        mu = numpy.arange(10)
        spectrum = Spectrum(energy=energy, mu=mu)
        numpy.testing.assert_array_equal(spectrum.energy, (energy * ur.eV).m)
        numpy.testing.assert_array_equal(spectrum.mu, mu)
        mu_2 = numpy.arange(30, 40)
        spectrum["Mu"] = mu_2
        numpy.testing.assert_array_equal(spectrum.mu, mu_2)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestSpectrum,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



import argparse
import sys
import logging
from est.core.types import XASObject
from est.core.reprocessing import get_process_instance_frm_h5_desc
from typing import Union

try:
    import h5py

    has_h5py = True
except:
    has_h5py = False

logging.basicConfig(level=logging.DEBUG)
_logger = logging.getLogger(__name__)


def exec_(
    h5_file: str,
    entry: str = "scan1",
    spectra_path: Union[None, str] = None,
    energy_path: Union[None, str] = None,
    configuration_path: Union[None, str] = None,
):

    try:
        xas_obj = XASObject.from_file(
            h5_file=h5_file,
            entry=entry or "scan1",
            spectra_path=spectra_path or "data/absorbed_beam",
            energy_path=energy_path or "data/energy",
            configuration_path=configuration_path,
        )
        xas_obj.link_to_h5(h5_file=h5_file)
    except Exception as e:
        _logger.error(e)
        xas_obj = None

    if not xas_obj:
        _logger.error("Unable to create an object from the given file and path")

    xas_processes = xas_obj.get_process_flow()
    if len(xas_processes) == 0:
        _logger.warning("no xas process found in the given h5 file")
        return
    xas_obj.clean_process_flow()
    for i_process, process in xas_processes.items():
        process_ins = get_process_instance_frm_h5_desc(process)
        xas_obj = process_ins(xas_obj)

    return xas_obj


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "workflow_file", help="Workflow file providing the workflow description (.h5)"
    )
    parser.add_argument(
        "--spectra_path",
        dest="spectra_path",
        default=None,
        help="path to the spectra data",
    )
    parser.add_argument(
        "--energy_path",
        dest="energy_path",
        default=None,
        help="path to the energy data",
    )
    parser.add_argument(
        "--configuration_path",
        dest="configuration_path",
        default=None,
        help="path to the process configuration",
    )
    parser.add_argument(
        "--entry",
        dest="entry to load",
        default=None,
        help="path to the process configuration",
    )
    options = parser.parse_args(argv[1:])
    exec_(
        h5_file=options.workflow_file,
        spectra_path=options.spectra_path,
        energy_path=options.energy_path,
        configuration_path=options.configuration_path,
    )


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/25/2019"

import os
import unittest
import tempfile
import shutil
import urllib.request
from est.core.types import Spectrum, XASObject

try:
    import larch
except ImportError:
    has_larch = False
else:
    has_larch = True
    from est.core.process.larch.mback import larch_mback, process_spectr_mback
    from est.io.utils.larch import read_ascii


@unittest.skipIf(has_larch is False, "xraylarch not installed")
class TestLarchSpectrum(unittest.TestCase):
    """Make sure computation on one spectrum is valid"""

    def setUp(self):
        self.outputdir = tempfile.mkdtemp()
        xmu_url = "https://raw.githubusercontent.com/xraypy/xraylarch/master/examples/xafs/cu_rt01.xmu"
        self.data_file = os.path.join(self.outputdir, "cu_rt01.xmu")

        with urllib.request.urlopen(xmu_url) as response, open(
            self.data_file, "wb"
        ) as out_file:
            data = response.read()  # a `bytes` object
            out_file.write(data)

        assert os.path.exists(self.data_file)
        energy, mu = read_ascii(self.data_file)
        self.spectrum = Spectrum(energy=energy, mu=mu)
        self.configuration = {"mback": {"z": 29}}

    def tearDown(self):
        shutil.rmtree(self.outputdir)

    def testProcess(self):
        self.assertFalse(hasattr(self.spectrum, "fpp"))
        self.assertFalse(hasattr(self.spectrum, "f2"))
        conf, res_spectrum = process_spectr_mback(self.spectrum, self.configuration)
        self.assertTrue(hasattr(res_spectrum, "fpp"))
        self.assertTrue(hasattr(res_spectrum, "f2"))


@unittest.skipIf(has_larch is False, "xraylarch not installed")
class TestLarchSpectra(unittest.TestCase):
    """Make sure computation on spectra is valid (n spectrum)"""

    def setUp(self):
        self.outputdir = tempfile.mkdtemp()
        # Download test file from xraylarch
        xmu_url = "https://raw.githubusercontent.com/xraypy/xraylarch/master/examples/xafs/cu_rt01.xmu"
        self.data_file = os.path.join(self.outputdir, "cu_rt01.xmu")
        with urllib.request.urlopen(xmu_url) as response, open(
            self.data_file, "wb"
        ) as out_file:
            data = response.read()  # a `bytes` object
            out_file.write(data)
        assert os.path.exists(self.data_file)
        # define the xas object
        self.configuration = {"mback": {"z": 29}}
        energy, mu = read_ascii(self.data_file)
        self.xas_object = XASObject(
            spectra=(Spectrum(energy=energy, mu=mu),),
            energy=energy,
            dim1=1,
            dim2=1,
            configuration=self.configuration,
        )

    def tearDown(self):
        shutil.rmtree(self.outputdir)

    def testProcessXASObject(self):
        res = larch_mback(self.xas_object)
        assert isinstance(res, XASObject)
        spectrum0 = res.spectra.data.flat[0]
        self.assertTrue(hasattr(spectrum0, "fpp"))
        self.assertTrue(hasattr(spectrum0, "f2"))

    def testProcessAsDict(self):
        res = larch_mback(self.xas_object.to_dict())
        assert isinstance(res, XASObject)
        spectrum0 = res.spectra.data.flat[0]
        self.assertTrue(hasattr(spectrum0, "fpp"))
        self.assertTrue(hasattr(spectrum0, "f2"))


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestLarchSpectrum, TestLarchSpectra):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Define some utils relative to larch"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "02/12/2019"


from larch.symboltable import Group


def group_to_dict(group):
    """Convert the larch group to a serializable dictionary

    :param group: the group to convert to a serializable dictionary
    :type: larch.symboltable.Group
    :returns: dictionary corresponding to the given larch.symboltable.Group
    :rtype: dictionary
    """
    res = {}
    for key in group._members():
        if isinstance(group._members()[key], Group):
            res[key] = group_to_dict(group._members()[key])
        else:
            res[key] = group._members()[key]


def dict_to_group(dict_, group):
    """Update the given larch group with the content of the dictionary

    :param dict_:
    :type: dict
    :param group:
    :type: larch.symboltable.Group
    """
    for key in dict_:
        group._members()[key] = dict_[key]



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"


from Orange.widgets import gui
from Orange.widgets.widget import OWWidget
from Orange.widgets.settings import Setting
from Orange.widgets.widget import Input
from silx.gui import qt
from est.core.types import XASObject
import est.core.io
import logging
import h5py

_logger = logging.getLogger(__file__)


class XASOutputOW(OWWidget):
    """
    Widget used for signal extraction
    """

    name = "xas output"
    id = "orange.widgets.xas.utils.xas_output"
    description = "Store process result (configuration)"
    icon = "icons/output.png"
    priority = 5
    category = "esrfWidgets"
    keywords = ["spectroscopy", "signal", "output", "file"]

    want_main_area = True
    resizing_enabled = True
    allows_cycle = False

    _output_file_setting = Setting(str())
    process_function = est.core.io.XASWriter

    class Inputs:
        xas_obj = Input("xas_obj", XASObject, default=True)

    def __init__(self):
        super().__init__()
        self._outputWindow = qt.QWidget(parent=self)
        self._outputWindow.setLayout(qt.QGridLayout())

        self._outputWindow.layout().addWidget(qt.QLabel("file", parent=self))
        self._inputLe = qt.QLineEdit("", parent=self)
        self._outputWindow.layout().addWidget(self._inputLe, 0, 0)
        self._selectPB = qt.QPushButton("select", parent=self)
        self._outputWindow.layout().addWidget(self._selectPB, 0, 1)

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self._outputWindow.layout().addWidget(spacer, 2, 0)

        layout = gui.vBox(self.mainArea, "output").layout()
        layout.addWidget(self._outputWindow)

        # deal with settings
        self.setFileSelected(self._output_file_setting)

        # signal / slot connection
        self._selectPB.pressed.connect(self._selectFile)

    def _selectFile(self, *args, **kwargs):
        old = self.blockSignals(True)
        dialog = qt.QFileDialog(self)
        # dialog.setFileMode(qt.QFileDialog.AnyFile)
        dialog.setAcceptMode(qt.QFileDialog.AcceptSave)
        dialog.setNameFilters(["hdf5 files (*.hdf5, *.hdf, *.h5)"])

        if not dialog.exec_():
            dialog.close()
            return False

        fileSelected = dialog.selectedFiles()
        if len(fileSelected) == 0:
            return False
        else:
            assert len(fileSelected) == 1
            file_ = fileSelected[0]
            if not h5py.is_hdf5(file_):
                file_ += ".h5"
            self.setFileSelected(file_)

        self.blockSignals(old)
        return True

    def setFileSelected(self, file_path):
        self._output_file_setting = file_path
        self._inputLe.setText(file_path)

    def _getFileSelected(self):
        return self._inputLe.text()

    @Inputs.xas_obj
    def process(self, xas_obj):
        if xas_obj is None:
            return
        if isinstance(xas_obj, dict):
            _xas_obj = XASObject.from_dict(xas_obj)
        else:
            _xas_obj = xas_obj

        has_file = self._getFileSelected() != ""
        if not has_file:
            mess = qt.QMessageBox(parent=self)
            mess.setIcon(qt.QMessageBox.Warning)
            mess.setText("No output file defined, please give a file path")
            mess.exec_()
            has_file = self._selectFile()
        if not has_file:
            _logger.error("no valid output file given, skip save")
        else:
            _xas_obj.dump(self._getFileSelected())



aliases = {
    # define aliases for larch
    "orangecontrib.est.widgets.larch.autobk.AutobkOW": "est.core.process.larch.autobk.Larch_autobk",
    "orangecontrib.est.widgets.larch.mback.MbackOW": "est.core.process.larch.mback.Larch_mback",
    "orangecontrib.est.widgets.larch.mback_norm.Mback_normOW": "est.core.process.larch.mback_norm.Larch_mback_norm",
    "orangecontrib.est.widgets.larch.pre_edge.PreEdgeOW": "est.core.process.larch.pre_edge.Larch_pre_edge",
    "orangecontrib.est.widgets.larch.xftf.XFTFOW": "est.core.process.larch.xftf.Larch_xftf",
    # define aliases for pymca
    "orangecontrib.est.widgets.pymca.exafs.ExafsOW": "est.core.process.pymca.exafs.PyMca_exafs",
    "orangecontrib.est.widgets.pymca.ft.FTOW": "est.core.process.pymca.ft.PyMca_ft",
    "orangecontrib.est.widgets.pymca.k_weight.KWeightOW": "est.core.process.pymca.k_weight.PyMca_k_weight",
    "orangecontrib.est.widgets.pymca.normalization.NormalizationOW": "est.core.process.pymca.normalization.PyMca_normalization",
    # define aliases for utils
    "orangecontrib.est.widgets.utils.roi.RoiSelectionOW": "est.core.process.roi.ROIProcess",
    "orangecontrib.est.widgets.utils.converter.ConverterOW": "est.core.process.reconstruction.axis.AxisProcess",
    "orangecontrib.est.widgets.utils.e0calculator.E0calculatorOW": "pypushflow.utils.IgnoreProcess",
    "orangecontrib.est.widgets.utils.noise.NoiseOW": "est.core.process.noise.NoiseProcess",
    "orangecontrib.est.widgets.utils.saving_points.SavingPointOW": "pypushflow.utils.IgnoreProcess",
    "orangecontrib.est.widgets.utils.xas_input.XASInputOW": "pypushflow.utils.IgnoreProcess",
    "orangecontrib.est.widgets.utils.xas_output.XASOutputOW": "est.core.process.io.DumpXasObject",
    "orangecontrib.est.widgets.utils.energyroi.EnergyRoiOW": "est.core.process.energyroi.EnergyROIProcess",
}



# coding: utf-8
###########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################

import sysconfig

# Category metadata.

# Category icon show in the menu
ICON = "icons/icon.png"

# Background color for category background in menu
# and widget icon background in workflow.
BACKGROUND = "light-green"

URL = "https://gitlab.esrf.fr/workflow/est"

WIDGET_HELP_PATH = (
    # Used for development.
    # You still need to build help pages using
    # make htmlhelp
    # inside doc folder
    ("{DEVELOP_ROOT}/build/sphinx/htmlhelp/index.html", None),
    #
    # Documentation included in wheel
    # Correct DATA_FILES entry is needed in setup.py and documentation has to be built
    # # before the wheel is created.
    ("{}/help/est/index.html".format(sysconfig.get_path("data")), None),
    ("http://www.silx.org/pub/doc/est/.htmlhelp/latest/", ""),
)


# Entry point for main Orange categories/widgets discovery
def widget_discovery(discovery):
    import pkg_resources

    dist = pkg_resources.get_distribution("est")
    pkgs = [
        "orangecontrib.est.widgets.pymca",
        "orangecontrib.est.widgets.larch",
        "orangecontrib.est.widgets.utils",
    ]
    for pkg in pkgs:
        discovery.process_category_package(pkg, distribution=dist)


import unittest


def suite():
    test_suite = unittest.TestSuite()

    from .test import suite as widgets_suite

    test_suite.addTest(widgets_suite())
    return test_suite



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""
unit test for workflow composed of pymca process 
"""
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/11/2019"


import os
import shutil
import tempfile
import unittest
import h5py
import numpy
from silx.io.utils import h5py_read_dataset
from silx.io.dictdump import h5todict
from silx.io.url import DataUrl
from est.core.process.roi import xas_roi, _ROI as XASROI
from est.core.io import XASWriter
from est.core.io import read as read_xas
from est.core.types import XASObject
from est.core.utils import spectra as spectra_utils
from est.core.types import Dim

try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
    from est.core.process.pymca.exafs import pymca_exafs, PyMca_exafs
    from est.core.process.pymca.ft import pymca_ft, PyMca_ft
    from est.core.process.pymca.k_weight import pymca_k_weight, PyMca_k_weight
    from est.core.process.pymca.normalization import pymca_normalization
    from est.core.process.pymca.normalization import PyMca_normalization


@unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
class TestStreamSingleSpectrum(unittest.TestCase):
    """Make sure the process have valid io"""

    def setUp(self):
        self.exafs_configuration = {
            "Knots": {"Values": (1, 2, 5), "Number": 3, "Orders": [3, 3, 3]},
            "KMin": 0,
            "KMax": 2.3,
        }

    def test_pymca_process(self):
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        out = read_xas(
            spectra_url=DataUrl(
                file_path=data_file, scheme="PyMca", data_path="Column 2"
            ),
            channel_url=DataUrl(
                file_path=data_file, scheme="PyMca", data_path="Column 1"
            ),
        )
        out.configuration = {"EXAFS": self.exafs_configuration, "SET_KWEIGHT": 0}
        out = pymca_normalization(xas_obj=out)
        out = pymca_exafs(xas_obj=out)
        out = pymca_k_weight(xas_obj=out)
        out = pymca_ft(xas_obj=out)
        assert isinstance(out, XASObject)

    def test_pymca_process_with_dict(self):
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        out = read_xas(
            spectra_url=DataUrl(
                file_path=data_file, scheme="PyMca", data_path="Column 2"
            ),
            channel_url=DataUrl(
                file_path=data_file, scheme="PyMca", data_path="Column 1"
            ),
        )
        out.configuration = {"EXAFS": self.exafs_configuration, "SET_KWEIGHT": 0}
        out = pymca_normalization(xas_obj=out.to_dict())
        out = pymca_exafs(xas_obj=out.to_dict())
        out = pymca_k_weight(xas_obj=out.to_dict())
        out = pymca_ft(xas_obj=out.to_dict())
        assert isinstance(out, XASObject)

    def test_pymca_process_with_cls(self):
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        out = read_xas(
            spectra_url=DataUrl(
                file_path=data_file, scheme="PyMca", data_path="Column 2"
            ),
            channel_url=DataUrl(
                file_path=data_file, scheme="PyMca", data_path="Column 1"
            ),
        )
        out = PyMca_normalization()(xas_obj=out)
        exafs_process = PyMca_exafs()
        exafs_process.setConfiguration(configuration=self.exafs_configuration)
        out = exafs_process(xas_obj=out)
        kweight_process = PyMca_k_weight()
        kweight_process.setConfiguration({"k_weight": 0})
        out = kweight_process(xas_obj=out)
        out = PyMca_ft()(xas_obj=out)
        assert isinstance(out, XASObject)
        self.assertTrue(out.spectra.data.flat[0].ft is not None)
        self.assertTrue(len(out.spectra.data.flat[0].ft.intensity) > 0)

    def test_pymca_with_roi(self):
        energy, spectra = spectra_utils.create_dataset(shape=(256, 20, 10))
        roi = XASROI(origin=(0, 2), size=(5, 1))
        xas_obj = XASObject(
            spectra=spectra, energy=energy, configuration={"roi": roi.to_dict()}
        )
        dict_xas_obj = xas_obj.to_dict()
        self.assertTrue("spectra" in dict_xas_obj.keys())
        self.assertTrue("energy" in dict_xas_obj.keys())
        tmp_obj = XASObject.from_dict(dict_xas_obj)
        numpy.testing.assert_array_equal(
            tmp_obj.energy, tmp_obj.spectra.data.flat[0].energy
        )
        out = xas_roi(dict_xas_obj)
        out.configuration = {"EXAFS": self.exafs_configuration, "SET_KWEIGHT": 0}
        out = pymca_normalization(xas_obj=out)
        out = pymca_exafs(xas_obj=out)
        out = pymca_k_weight(xas_obj=out)
        out = pymca_ft(xas_obj=out)
        assert isinstance(out, XASObject)
        self.assertEqual(out.spectra.shape[0], 1)
        self.assertEqual(out.spectra.shape[1], 5)
        self.assertTrue(out.spectra.data.flat[0].ft is not None)
        self.assertTrue(len(out.spectra.data.flat[0].ft.intensity) > 0)


@unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
class TestWorkflowAndH5LinkedFile(unittest.TestCase):
    """
    Test that the workflow can process and store process in a targeted h5 file
    """

    def setUp(self):
        self.output_dir = tempfile.mkdtemp()
        energy, spectra = spectra_utils.create_dataset(shape=(256, 20, 10))
        roi = XASROI(origin=(0, 2), size=(5, 1))
        self.xas_obj = XASObject(
            spectra=spectra, energy=energy, configuration={"roi": roi.to_dict()}
        )
        self.h5_file = os.path.join(self.output_dir, "output_file.h5")
        self.xas_obj.link_to_h5(self.h5_file)

    def tearDown(self):
        shutil.rmtree(self.output_dir)

    def test_h5_link_xas_object(self):
        """Test that the processing can be stored continuously on a .h5 file"""
        self.assertTrue(self.xas_obj.linked_h5_file is not None)
        out = PyMca_normalization()(xas_obj=self.xas_obj)
        exafs_process = PyMca_exafs()
        configuration = {
            "Knots": {"Values": (1, 2, 5), "Number": 3, "Orders": [3, 3, 3]},
            "KMin": 0,
            "KMax": 2.3,
        }
        exafs_process.setConfiguration(configuration=configuration)
        out = exafs_process(xas_obj=out)
        k_weight_process = PyMca_k_weight()
        k_weight_process.setConfiguration({"k_weight": 0})
        out = k_weight_process(xas_obj=out)
        out = PyMca_ft()(xas_obj=out)
        out = PyMca_normalization()(xas_obj=out)
        assert isinstance(out, XASObject)
        assert out.linked_h5_file is self.h5_file
        # then check all process are correctly registered with the valid id...
        processes = self.xas_obj.get_process_flow()
        self.assertEqual(len(processes), 5)
        self.assertEqual(
            h5py_read_dataset(processes[1]["program"]), "pymca_normalization"
        )
        self.assertEqual(h5py_read_dataset(processes[2]["program"]), "pymca_exafs")
        self.assertEqual(
            h5py_read_dataset(processes[5]["program"]), "pymca_normalization"
        )
        self.xas_obj.clean_process_flow()
        processes = self.xas_obj.get_process_flow()
        self.assertEqual(len(processes), 0)

    def test_h5_link_dict(self):
        """Same test as test_h5_link_xas_object but with a dict pass between
        processes"""
        self.assertTrue(self.xas_obj.linked_h5_file is not None)
        out = PyMca_normalization()(xas_obj=self.xas_obj.to_dict())
        exafs_process = PyMca_exafs()
        configuration = {
            "Knots": {"Values": (1, 2, 5), "Number": 3, "Orders": [3, 3, 3]},
            "KMin": 0,
            "KMax": 2.3,
        }
        exafs_process.setConfiguration(configuration=configuration)
        out = exafs_process(xas_obj=out.to_dict())
        k_weight_process = PyMca_k_weight()
        k_weight_process.setConfiguration({"k_weight": 0})
        out = k_weight_process(xas_obj=out.to_dict())
        out = PyMca_ft()(xas_obj=out.to_dict())
        out = PyMca_normalization()(xas_obj=out.to_dict())
        assert isinstance(out, XASObject)
        assert out.linked_h5_file
        assert out.linked_h5_file == self.h5_file
        # then check all process are correctly registered with the valid id...
        processes = self.xas_obj.get_process_flow()
        self.assertEqual(len(processes), 5)
        self.assertEqual(
            h5py_read_dataset(processes[1]["program"]), "pymca_normalization"
        )
        self.assertEqual(h5py_read_dataset(processes[2]["program"]), "pymca_exafs")
        self.assertEqual(
            h5py_read_dataset(processes[5]["program"]), "pymca_normalization"
        )
        self.xas_obj.clean_process_flow()
        processes = self.xas_obj.get_process_flow()
        self.assertEqual(len(processes), 0)


@unittest.skipIf(has_pymca is False, "PyMca5 is not installed")
class TestSaveFlowAuto(unittest.TestCase):
    """Test that the processing can be stored continuously on a temporary
    .h5 file then dump without lost during call to 'dump_xas' of the XASObject
    """

    def setUp(self):
        self.output_dir = tempfile.mkdtemp()
        energy, spectra = spectra_utils.create_dataset(shape=(100, 10, 10))
        assert spectra.shape == (100, 10, 10)
        assert len(energy) == spectra.shape[0]
        self.xas_obj = XASObject(spectra=spectra, energy=energy, dim1=10, dim2=10)

        self.h5_file = os.path.join(self.output_dir, "output_file.h5")

    def tearDown(self):
        shutil.rmtree(self.output_dir)

    def test(self):
        self.assertTrue(self.xas_obj.linked_h5_file is not None)
        out = PyMca_normalization()(xas_obj=self.xas_obj)
        exafs = PyMca_exafs()
        configuration = {
            "Knots": {"Values": (1, 2, 5), "Number": 3, "Orders": [3, 3, 3]},
            "KMin": 0,
            "KMax": 2.3,
        }
        exafs.setConfiguration(configuration)
        out = exafs(xas_obj=out)
        kweight_process = PyMca_k_weight()
        kweight_process.setConfiguration({"k_weight": 0})
        out = kweight_process(xas_obj=out)
        out = PyMca_ft()(xas_obj=out)
        out = PyMca_normalization()(xas_obj=out)

        writer = XASWriter()
        writer.output_file = self.h5_file
        writer(out)

        with h5py.File(self.h5_file, "r") as hdf:
            self.assertTrue("scan1" in hdf.keys())
            self.assertTrue("data" in hdf["scan1"].keys())
            self.assertTrue("absorbed_beam" in hdf["scan1"].keys())
            self.assertTrue("monochromator" in hdf["scan1"].keys())
            self.assertTrue("xas_process_1" in hdf["scan1"].keys())
            self.assertTrue("xas_process_2" in hdf["scan1"].keys())
            self.assertTrue("xas_process_3" in hdf["scan1"].keys())
            self.assertTrue("xas_process_4" in hdf["scan1"].keys())
            self.assertTrue("xas_process_5" in hdf["scan1"].keys())
            targetted_xas_process = hdf["scan1"]["xas_process_2"]
            self.assertTrue("program" in targetted_xas_process)
            self.assertEqual(
                h5py_read_dataset(targetted_xas_process["program"]), "pymca_exafs"
            )
            self.assertTrue("configuration" in targetted_xas_process)

        # check one configuration
        stored_config = h5todict(
            self.h5_file, path="/scan1/xas_process_2/configuration", asarray=False
        )
        for key in ("KMin", "KMax"):
            self.assertTrue(configuration[key] == stored_config[key])


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestStreamSingleSpectrum, TestWorkflowAndH5LinkedFile, TestSaveFlowAuto):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/12/2020"


import pint
from typing import Iterable
from typing import Union
import numpy
from est.units import ur
from est.units import convert_to as convert_unit_to
from silx.io.url import DataUrl
from .spectrum import Spectrum

try:
    import larch.symboltable
except ImportError:
    _has_larch = False
else:
    _has_larch = True
import logging

_logger = logging.getLogger(__name__)


class Spectra:
    """
    A map of spectrum.

    - data: numpy array of Spectrum. Expected to be 2D.
    - energy: set of energy for each position (x,y)
    """

    def __init__(self, energy, spectra: Union[Iterable, None] = None):
        self.__energy = energy
        if spectra is None:
            spectra = []
        if not isinstance(spectra, (list, tuple, numpy.ndarray)):
            raise TypeError("spectra type is invalid: {}".format(type(spectra)))

        if energy is not None:
            self.__energy = convert_unit_to(energy, ur.eV).m_as(ur.eV)
        else:
            self.__energy = None

        spectrum_list = []
        shape = (1, -1)
        if isinstance(spectra, numpy.ndarray):
            if not spectra.ndim == 3:
                raise ValueError("Provided spectra is expected to be 3d")

            for y_i_spectrum in range(spectra.shape[1]):
                for x_i_spectrum in range(spectra.shape[2]):
                    spectrum_list.append(
                        Spectrum(
                            energy=energy, mu=spectra[:, y_i_spectrum, x_i_spectrum]
                        )
                    )
            # spectra dimensions should be channel, y, x
            shape = spectra.shape[-2:]
        else:
            for spectrum in spectra:
                assert isinstance(spectrum, Spectrum)
                spectrum_list.append(spectrum)
        self.__data = numpy.array(spectrum_list)
        for spec in self.__data.flat:
            assert isinstance(spec, Spectrum)
        self.reshape(shape)

    def check_validity(self):
        for spectrum in self.data.flat:
            assert isinstance(spectrum, Spectrum)
            assert isinstance(spectrum.energy, (numpy.ndarray, pint.Quantity))
            assert isinstance(spectrum.mu, numpy.ndarray)

    @property
    def data(self):
        return self.__data

    def __getitem__(self, item):
        return self.__data[item]

    @property
    def shape(self):
        return self.__data.shape

    def reshape(self, shape):
        assert len(shape) == 2
        assert isinstance(shape, Iterable)
        if None in shape:
            raise ValueError("None not handled")
        self.__data = self.__data.reshape(shape)
        for spec in self.__data.flat:
            assert isinstance(spec, Spectrum)
        return self.__data

    @property
    def energy(self):
        return self.__energy

    @property
    def energy_with_unit(self) -> pint.Quantity:
        return self.__energy * ur.eV

    @energy.setter
    def energy(self, energy):
        if isinstance(energy, numpy.ndarray):
            energy = energy * ur.eV
        if energy is None:
            self.__energy = None
        else:
            self.__energy = energy.m_as(ur.eV)
        if len(self.data) > 0:
            if len(self.data.flat[0].energy) != len(energy):
                _logger.warning("spectra and energy have incoherent dimension")

    def map_to(self, data_info: str, relative_to: str = "energy"):
        """
        Create a map a specific key of this spectra.
        """

        def get_param(object, name):
            if hasattr(object, name):
                return getattr(object, name)
            else:
                return object[name]

        def get_relative_to_value(relative):
            paths_ = relative.split(".")
            value = getattr(self.data.flat[0], paths_[0])
            for path_ in paths_[1:]:
                value = getattr(value, path_)
            return value

        if len(self.data.flat) == 0:
            return None
        else:

            def allocate_array(relative_to_len):
                if relative_to_len is not None:
                    return numpy.zeros((relative_to_len, len(self.data.flat)))
                else:
                    return numpy.zeros(len(self.data.flat))

            key = str(data_info)
            array = None
            for i_spectrum, spectrum in enumerate(self.data.flat):
                try:
                    if "." in key:
                        subkeys = key.split(".")
                        key_ = subkeys[-1]
                        subkeys = subkeys[:-1]
                        value = get_param(spectrum, subkeys[0])
                        for subkey in subkeys[1:]:
                            value = get_param(value, subkey)
                        value = get_param(value, key_)
                    else:
                        value = get_param(spectrum, key)
                except Exception:
                    _logger.info("fail to access to {}".format(key))
                    break
                else:
                    if isinstance(value, pint.Quantity):
                        value = convert_unit_to(value, ur.eV).m_as(ur.eV)
                    if _has_larch and isinstance(value, larch.symboltable.Group):
                        _logger.info("pass larch details, not managed for now")
                        continue
                    # create array if necessary
                    if array is None:
                        if relative_to is None:
                            array = allocate_array(relative_to_len=None)
                        elif value is None:
                            array = allocate_array(relative_to_len=None)
                        else:
                            array = allocate_array(relative_to_len=len(value))
                    if relative_to is not None:
                        array[:, i_spectrum] = value
                    else:
                        array[i_spectrum] = value
            shape = list(self.data.shape)
            shape.insert(0, -1)
            if array is None:
                return array
            else:
                return array.reshape(shape)

    def keys(self) -> tuple:
        """
        return the tuple of keys contained in the spectra
        """
        if len(self.data.flat) > 0:
            assert isinstance(self.data.flat[0], Spectrum)
            return tuple(self.data.flat[0].keys())

    def __eq__(self, other):
        if not isinstance(other, Spectra):
            return False
        if len(self.data) != len(other.data):
            return False
        else:
            return numpy.array_equal(self.data.flat, other.data.flat)

    def __iter__(self):
        return iter(self.data.flat)

    @staticmethod
    def frm_dict(ddict, dimensions):
        """
        :param dict ddict: dict containing the data to be loaded
        :param tuple dimensions: information regarding spectra dimensions
        """
        from est.io import load_data  # avoid cyclic import

        # if spectra is given from an url
        if isinstance(ddict, str):
            return load_data(
                data_url=DataUrl(path=ddict),
                name="spectra",
                dimensions=dimensions,
            )
        # if come from a list of spectrum
        elif not isinstance(ddict, (numpy.ndarray, pint.Quantity)):
            new_spectra = []
            for spectrum in ddict:
                assert isinstance(spectrum, dict)
                new_spectra.append(Spectrum.from_dict(spectrum))
            return Spectra(energy=new_spectra[0].energy, spectra=new_spectra)
        else:
            raise TypeError("Unhandled input type ({})".format(type(ddict)))



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""Tools to select energy unit"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "31/03/2020"


from silx.gui import qt
from est.units import ur


class EnergyUnitSelector(qt.QComboBox):
    """Simple class to define a unit for energy"""

    def __init__(self, parent=None):
        qt.QComboBox.__init__(self, parent)
        for unit in ("eV", "keV", "J", "kJ"):
            self.addItem(unit)

    def getUnit(self):
        current_unit = self.currentText()
        if current_unit == "eV":
            return ur.eV
        elif current_unit == "keV":
            return ur.keV
        elif current_unit == "J":
            return ur.J
        elif current_unit == "kJ":
            return ur.kJ
        else:
            raise ValueError("current unit is not supported")

    def setUnit(self, unit: str) -> None:
        if unit in ("eV", "keV", "J", "kJ"):
            txt = unit
        elif unit == ur.eV:
            txt = "eV"
        elif unit == ur.keV:
            txt = "keV"
        elif unit == ur.J:
            txt = "J"
        elif unit == ur.kJ:
            txt = "kJ"
        else:
            raise ValueError("Given unit is not managed: {}".format())

        index = self.findText(txt)
        if index >= 0:
            self.setCurrentIndex(index)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/12/2019"

import logging
from datetime import datetime
import h5py
import numpy
from silx.io.dictdump import dicttoh5, h5todict
from silx.io.url import DataUrl
from silx.utils.enum import Enum
from est.units import ur
from est.core.types.dim import Dim
from est.core.types import Spectra
from est.io.utils.spec import read_spectrum as read_spec_spectrum
from est.thirdparty.silx.hdf5file import get_data, HDF5File

try:
    from est.io.utils.larch import read_ascii as larch_read_ascii
except ImportError:
    has_larch = False
else:
    has_larch = True

_logger = logging.getLogger(__name__)


class InputType(Enum):
    dat_spectrum = "*.dat"  # contains one spectrum
    hdf5_spectra = "*.h5"  # contains several spectra
    xmu_spectrum = "*.xmu"  # contains one spectrum
    csv_spectrum = "*.csv"  # two columns, comma separated


def move_axis_to_standard(spectra, dimensions):
    # make sure all dimensions are defined
    for dim in Dim:
        if dim not in dimensions:
            err = "%s is not defined in the dimensions" % dim
            raise ValueError(err)
    # fit spectra according to dimension
    src_axis = (
        dimensions.index(Dim.DIM_2),
        dimensions.index(Dim.DIM_1),
        dimensions.index(Dim.DIM_0),
    )
    dst_axis = (0, 1, 2)
    _logger.warning("move axis for spectra, from %s to %s" % (src_axis, dst_axis))
    if isinstance(spectra, Spectra):
        spectra.data = numpy.moveaxis(spectra.data, src_axis, dst_axis)
    elif isinstance(spectra, numpy.ndarray):
        spectra = numpy.moveaxis(spectra, src_axis, dst_axis)
    return spectra


def load_data(data_url, name, dimensions, columns_names=None, energy_unit=ur.eV):
    """
    Load a specific data from an url. Manage the different scheme (silx, fabio,
    numpy, PyMca, xraylarch)

    :param data_url: silx DataUrl with path to the data
    :type: DataUrl
    :param str name: name of the data we want to load. Should be in
                    ('spectra', 'energy', 'configuration')
    :param Union[None,dict] columns_names: name of the column to pick for .dat
                                           files... Expect key 'mu' and
                                                    'energy' to be registered
    :return: data loaded
    :rtype: Union[None,dict,numpy.ndarray]
    """
    if data_url is None:
        return None
    from est.core.types import Dim  # avoid cyclic import

    if dimensions is not None:
        dimensions = tuple([Dim.from_value(dim) for dim in dimensions])
    assert isinstance(data_url, DataUrl)
    if data_url.scheme() in ("PyMca", "PyMca5", "Spec", "spec"):

        def get_energy_col_name():
            if columns_names is not None:
                return columns_names["energy"]
            if name == "energy" and data_url.data_path() is not None:
                return data_url.data_path()
            else:
                return None

        def get_absorption_col_name():
            if columns_names is not None:
                return columns_names["mu"]
            if name == "spectra" and data_url.data_path() is not None:
                return data_url.data_path()
            else:
                return None

        energy, mu = read_spec_spectrum(
            data_url.file_path(),
            energy_col_name=get_energy_col_name(),
            absorption_col_name=get_absorption_col_name(),
            monitor_col_name=columns_names["monitor"] if columns_names else None,
            scan_header_S=columns_names["scan_title"] if columns_names else None,
            energy_unit=energy_unit,
        )
        if name == "spectra":
            return mu.reshape(-1, 1, 1)
        else:
            return energy
    elif data_url.scheme() in ("larch", "xraylarch"):
        if has_larch is False:
            _logger.warning("Requires larch to load data from " "%s" % data_url.path())
            return None
        else:
            assert name in ("spectra", "energy")
            energy, mu = larch_read_ascii(
                xmu_file=data_url.file_path(), energy_unit=energy_unit
            )

            if name == "spectra":
                mu = numpy.ascontiguousarray(mu[:])
                return mu.reshape(mu.shape[0], 1, 1)
            else:
                return energy
    elif data_url.scheme() == "numpy":
        return move_axis_to_standard(numpy.load(data_url.file_path()), dimensions)
    elif data_url.scheme() == "est":
        assert name == "spectra"
        spectra = []
        with HDF5File(data_url.file_path(), "r") as hdf5:
            # get all possible entries
            entries = filter(
                lambda x: isinstance(hdf5[x], h5py.Group)
                and "est_saving_pt" in hdf5[x].keys(),
                hdf5.keys(),
            )
            entries = list(entries)
            if len(entries) == 0:
                _logger.error(
                    "no spectra dataset found in the file", data_url.file_path()
                )
                return

            if len(entries) > 1:
                _logger.warning(
                    "several entry detected, only one will be loaded:", entries[0]
                )
            spectra_path = "/".join((entries[0], "est_saving_pt", "spectra"))
            node_spectra = hdf5[spectra_path]
            spectrum_indexes = list(node_spectra.keys())
            spectrum_indexes = list(map(lambda x: int(x), spectrum_indexes))
            spectrum_indexes.sort()
            from est.core.types import Spectrum

            for index in spectrum_indexes:
                spectrum_path = "/".join((spectra_path, str(index)))
                dict_ = h5todict(
                    h5file=data_url.file_path(), path=spectrum_path, asarray=False
                )
                spectrum = Spectrum().load_frm_dict(dict_)
                spectra.append(spectrum)
        return Spectra(energy=spectra[0].energy, spectra=spectra)
    else:
        if data_url.is_valid():
            try:
                data = get_data(data_url)
            except ValueError as e:
                _logger.error(e)
            else:
                if name == "spectra":
                    if data.ndim == 1:
                        return data.reshape(data.shape[0], 1, 1)
                    elif data.ndim == 3:
                        return move_axis_to_standard(data, dimensions=dimensions)
                return data
        else:
            _logger.warning(
                "invalid url for {}: {}  will not load it".format(name, data_url)
            )


def read_xas(information):
    """
    Read the given spectra, configuration... from the provided input
    Information

    :param InputInformation informationUnion:

    :return: spectra, energy, configuration
    """

    def get_url(original_url, name):
        url_ = original_url
        if type(url_) is str:
            try:
                url_ = DataUrl(path=url_)
            except Exception:
                url_ = DataUrl(file_path=url_, scheme="PyMca")

        if not isinstance(url_, DataUrl):
            raise TypeError("given input for, {} is invalid".format(name))
        return url_

    _spectra_url = get_url(original_url=information.spectra_url, name="spectra")
    _energy_url = get_url(original_url=information.channel_url, name="energy")
    _config_url = information.config_url
    if type(_config_url) is str and _config_url == "":
        _config_url = None
    if not (_config_url is None or isinstance(_config_url, DataUrl)):
        raise TypeError("given input for configuration is invalid")

    from est.core.types import Dim  # avoid cyclic import

    # this should be extractable and done in the InputInformation class
    dimensions_ = information.dimensions
    if dimensions_ is None:
        dimensions_ = (Dim.DIM_2, Dim.DIM_1, Dim.DIM_0)
    else:
        dimensions_ = []
        for dim in information.dimensions:
            dimensions_.append(Dim.from_value(dim))
    spectra = load_data(
        _spectra_url,
        name="spectra",
        dimensions=dimensions_,
        columns_names=information.columns_names,
        energy_unit=information.energy_unit,
    )
    energy = load_data(
        _energy_url,
        name="energy",
        dimensions=dimensions_,
        columns_names=information.columns_names,
        energy_unit=information.energy_unit,
    )
    configuration = load_data(
        _config_url,
        name="configuration",
        dimensions=dimensions_,
        columns_names=information.columns_names,
        energy_unit=information.energy_unit,
    )

    if energy is None:
        raise ValueError("Unable to load energy from {}".format(_energy_url))
    if not energy.ndim == 1:
        raise ValueError("Energy / channel is not 1D")
    if energy.shape[0] > spectra.shape[0]:
        energy = energy[: spectra.shape[0]]
        _logger.warning("energy has less value than spectra. Clip energy.")
    if not energy.shape[0] == spectra.shape[0]:
        err = "Energy / channel and spectra dim1 have incoherent length (%s vs %s)" % (
            energy.shape[0],
            spectra.shape[0],
        )
        raise ValueError(err)
    return spectra, energy * information.energy_unit, configuration


def write_xas_proc(
    h5_file,
    entry,
    process,
    results,
    processing_order,
    plots,
    data_path="/",
    overwrite=True,
):
    """
    Write a xas :class:`.Process` into .h5

    :param str h5_file: path to the hdf5 file
    :param str entry: entry name
    :param process: process executed
    :type: :class:`.Process`
    :param results: process result data
    :type: numpy.ndarray
    :param processing_order: processing order of treatment
    :type: int
    :param data_path: path to store the data
    :type: str
    """
    if plots is None:
        plots = tuple()
    process_name = "xas_process_" + str(processing_order)
    # write the xasproc default information
    with HDF5File(h5_file, "a") as h5f:
        nx_entry = h5f.require_group("/".join((data_path, entry)))
        nx_entry.attrs["NX_class"] = "NXentry"

        nx_process = nx_entry.require_group(process_name)
        nx_process.attrs["NX_class"] = "NXprocess"
        if overwrite:
            for key in (
                "program",
                "version",
                "date",
                "processing_order",
                "class_instance",
                "ft",
            ):
                if key in nx_process:
                    del nx_process[key]
        nx_process["program"] = process.program_name()
        nx_process["version"] = process.program_version()
        nx_process["date"] = datetime.now().replace(microsecond=0).isoformat()
        nx_process["processing_order"] = numpy.int32(processing_order)
        _class = process.__class__
        nx_process["class_instance"] = ".".join((_class.__module__, _class.__name__))

        nx_data = nx_entry.require_group("data")
        nx_data.attrs["NX_class"] = "NXdata"
        nx_data.attrs["signal"] = "data"
        nx_process_path = nx_process.name

    if isinstance(results, numpy.ndarray):
        data_ = {"data": results}
    else:
        data_ = results

    def get_interpretation(my_data):
        """Return hdf5 attribute for this type of data"""
        if isinstance(my_data, numpy.ndarray):
            if my_data.ndim == 1:
                return "spectrum"
            elif my_data.ndim in (2, 3):
                return "image"
        return None

    def get_path_to_result(res_name):
        res_name = res_name.replace(".", "/")
        path = "/".join((entry, process_name, "results", res_name))
        if path.startswith("/"):
            path = "/" + path
        return path

    # save results
    def save_key(key_path, value, attrs):
        """Save the given value to the associated path. Manage numpy arrays
        and dictionaries.
        """
        if attrs is not None:
            assert value is None, "can save value or attribute not both"
        if value is not None:
            assert attrs is None, "can save value or attribute not both"
        key_path = key_path.replace(".", "/")
        # save if is dict
        if isinstance(value, dict):
            h5_path = "/".join((entry, process_name, key_path))
            dicttoh5(
                value,
                h5file=h5_file,
                h5path=h5_path,
                overwrite_data=True,
                mode="a",
            )
        else:
            with HDF5File(h5_file, "a") as h5f:
                nx_process = h5f.require_group(nx_process_path)
                if attrs is None:
                    try:
                        nx_process[key_path] = value
                    except TypeError as e:
                        _logger.warning(
                            "Unable to write at {} reason is {}"
                            "".format(str(key_path), str(e))
                        )
                    else:
                        interpretation = get_interpretation(value)
                        if interpretation:
                            nx_process[key_path].attrs[
                                "interpretation"
                            ] = interpretation
                else:
                    for key, value in attrs.items():
                        try:
                            nx_process[key_path].attrs[key] = value
                        except Exception as e:
                            _logger.warning(e)

    for key, value in data_.items():
        if isinstance(key, tuple):
            key_path = "/".join(("results", key[0]))
            save_key(key_path=key_path, value=None, attrs={key[1]: value})
        else:
            key_path = "/".join(("results", str(key)))
            save_key(key_path=key_path, value=value, attrs=None)

    def save_plot(plot_name, plot):
        """save the given plot to an hdf5 group"""
        plot_name = plot_name.replace(".", "/")
        plot_path = "/".join((entry, process_name, "plots", plot_name))

        with HDF5File(h5_file, "a") as h5f:
            plot_group = h5f.require_group(plot_path)
            plot_group.attrs["NX_class"] = "NXdata"
            plot_group.attrs["interpretation"] = "spectrum"
            assert plot.signal is not None
            assert plot.axes is not None

            def link_dataset(dataset_to_link, name):
                # to insure silx isplotting it we should have curve as a 1D object
                # but by default we are handling a map of spectra. This is why we
                # need to duplicate data here
                if dataset_to_link.ndim == 1:
                    plot_group[name] = h5py.SoftLink(dataset_to_link.name)
                elif dataset_to_link.ndim == 2:
                    plot_group[name] = dataset_to_link[:, 0]
                    for key in ("units", "units_latex"):
                        if key in dataset_to_link.attrs:
                            plot_group[name].attrs[key] = dataset_to_link.attrs[key]
                elif dataset_to_link.ndim == 3:
                    plot_group[name] = dataset_to_link[:, 0, 0]
                    for key in ("units", "units_latex"):
                        if key in dataset_to_link.attrs:
                            plot_group[name].attrs[key] = dataset_to_link.attrs[key]
                else:
                    raise ValueError(
                        "Unable to handle dataset {}".format(dataset_to_link.name)
                    )

            # handle signal
            # plot is only handling 1D data
            try:
                signal_dataset = h5f[get_path_to_result(plot.signal)]
            except KeyError:
                _logger.info("{} is not available".format(plot.signal))
            else:
                link_dataset(dataset_to_link=signal_dataset, name=plot.signal)
            plot_group.attrs["signal"] = plot.signal
            # handle axes
            for axe in plot.axes:
                try:
                    axe_dataset = h5f[get_path_to_result(axe)]
                except KeyError:
                    _logger.info("{} is not available".format(axe))
                else:
                    link_dataset(dataset_to_link=axe_dataset, name=axe)
            plot_group.attrs["axes"] = plot.axes
            # handle auxiliary signals
            if plot.auxiliary_signals is not None:
                for aux_sig in plot.auxiliary_signals:
                    try:
                        aux_sig_dataset = h5f[get_path_to_result(aux_sig)]
                    except KeyError:
                        _logger.info("{} is not available".format(aux_sig))
                    else:
                        link_dataset(dataset_to_link=aux_sig_dataset, name=aux_sig)

                plot_group.attrs["auxiliary_signals"] = plot.auxiliary_signals
            # handle title(s)
            if plot.title is not None:
                plot_group.attrs["title"] = plot.title
            if plot.title_latex is not None:
                plot_group.attrs["title_latex"] = plot.title_latex

            # handle silx style
            if plot.silx_style is not None:
                import json

                plot_group.attrs["SILX_style"] = json.dumps(plot.silx_style)

    # save plots
    for i_plot, plot in enumerate(plots):
        plot_name = "plot_{}".format(i_plot)
        save_plot(plot_name=plot_name, plot=plot)

    # default plot will always be the first one
    if len(plots) > 0:
        plots_path = "/".join((entry, process_name, "plots"))
        with HDF5File(h5_file, "a") as h5f:
            plots_group = h5f.require_group(plots_path)
            plots_group.attrs["NX_class"] = "NXdata"
            plots_group.attrs["default"] = "plot_0"

    if process.getConfiguration() is not None:
        h5_path = "/".join((nx_process_path, "configuration"))
        dicttoh5(
            process.getConfiguration(),
            h5file=h5_file,
            h5path=h5_path,
            overwrite_data=True,
            mode="a",
        )


def write_xas(
    h5_file,
    entry,
    energy,
    mu,
    sample=None,
    start_time=None,
    data_path="/",
    title=None,
    definition=None,
    overwrite=True,
):
    """
    Write raw date in nexus format

    :param str h5_file: path to the hdf5 file
    :param str entry: entry name
    :param sample: definition of the sample
    :type: :class:`.Sample`
    :param energy: beam energy (1D)
    :type: numpy.ndarray
    :param mu: beam absorption (2D)
    :type: numpy.ndarray
    :param start_time:
    :param str data_path:
    :param str title: experiment title
    :param str definition: experiment definition
    """
    with HDF5File(h5_file, "w") as h5f:
        nx_entry = h5f.require_group("/".join((data_path, entry)))
        nx_entry.attrs["NX_class"] = "NXentry"

        # store energy
        nx_monochromator = nx_entry.require_group("monochromator")
        nx_monochromator.attrs["NX_class"] = "NXmonochromator"
        if overwrite and "energy" in nx_monochromator:
            del nx_monochromator["energy"]
        nx_monochromator["energy"] = energy
        nx_monochromator["energy"].attrs["interpretation"] = "spectrum"
        nx_monochromator["energy"].attrs["NX_class"] = "NXdata"
        nx_monochromator["energy"].attrs["unit"] = "eV"

        # store absorbed beam
        nx_absorbed_beam = nx_entry.require_group("absorbed_beam")
        nx_absorbed_beam.attrs["NX_class"] = "NXdetector"
        if overwrite and "data" in nx_absorbed_beam:
            del nx_absorbed_beam["data"]
        nx_absorbed_beam["data"] = mu
        nx_absorbed_beam["data"].attrs["interpretation"] = "image"
        nx_absorbed_beam["data"].attrs["NX_class"] = "NXdata"

        if sample:
            nx_sample = nx_entry.require_group("sample")
            nx_sample.attrs["NX_class"] = "NXsample"
            if overwrite and "name" in nx_sample:
                del nx_sample["name"]
            nx_sample["name"] = sample.name

        nx_data = nx_entry.require_group("data")
        nx_data.attrs["NX_class"] = "NXdata"
        # create some link on data
        if overwrite and "energy" in nx_data:
            del nx_data["energy"]
        nx_data["energy"] = h5py.SoftLink(nx_monochromator["energy"].name)
        if overwrite and "absorbed_beam" in nx_data:
            del nx_data["absorbed_beam"]
        nx_data["absorbed_beam"] = h5py.SoftLink(nx_absorbed_beam["data"].name)

        if start_time is not None:
            if overwrite and "start_time" in nx_entry:
                del nx_entry["start_time"]
            nx_entry["start_time"] = start_time
        if title is not None:
            if overwrite and "title" in nx_entry:
                del nx_entry["title"]
            nx_entry["title"] = title
        if definition is not None:
            if overwrite and "definition" in nx_entry:
                del nx_entry["definition"]
            nx_entry["definition"] = definition


def write_spectrum_saving_pt(h5_file, entry, obj, overwrite=True):
    """Save the current status of an est object

    :param str h5_file: path to the hdf5 file
    :param str entry: entry name
    :param obj: object to save.
    :param str obj_name: name of the object to store
    :param str data_path:
    """
    dicttoh5(obj, h5file=h5_file, h5path=entry, overwrite_data=True, mode="a")


def get_xasproc(h5_file, entry):
    """
    Return the list of all NXxasproc existing at the data_path level

    :param str h5_file: hdf5 file
    :param str entry: data location

    :return:
    :rtype: list
    """

    def copy_nx_xas_process(h5_group):
        """copy base information from nx_xas_process"""
        res = {}
        res["_h5py_path"] = h5_group.name
        relevant_keys = (
            "program",
            "version",
            "data",
            "parameters",
            "processing_order",
            "configuration",
            "class_instance",
            "plots",
        )
        from silx.io.dictdump import h5todict

        for key in h5_group.keys():
            # for now we don't want to copy the numpy array (data)
            if key in relevant_keys:
                if key == "configuration":
                    config_path = "/".join((h5_group.name, "configuration"))
                    res[key] = h5todict(h5_file, config_path, asarray=False)
                elif key == "plots":
                    plots_grp = h5_group["plots"]
                    res[key] = {}
                    for plot_key in plots_grp.keys():
                        res[key][plot_key] = dict(plots_grp[plot_key].attrs.items())
                else:
                    res[key] = h5_group[key][...]
        return res

    res = []
    with HDF5File(h5_file, "a") as h5f:
        try:
            root_group = h5f[entry]
        except KeyError:
            _logger.warning(entry + " does not exist in " + h5_file)
        else:
            for key in root_group.keys():
                elmt = root_group[key]
                if hasattr(elmt, "attrs") and "NX_class" in elmt.attrs:
                    if elmt.attrs["NX_class"] == "NXprocess":
                        nx_xas_proc = copy_nx_xas_process(elmt)
                        if len(nx_xas_proc) == 0:
                            _logger.warning(
                                "one xas process was not readable "
                                "from the hdf5 file at:" + key
                            )
                        else:
                            res.append(nx_xas_proc)
    return res


if __name__ == "__main__":
    import os
    from est.core.process.pymca.normalization import PyMca_normalization
    from est.core.process.pymca.exafs import PyMca_exafs
    from est.core.types import Sample

    h5_file = "test_xas_123.h5"
    if os.path.exists(h5_file):
        os.remove(h5_file)
    sample = Sample(name="mysample")
    data = numpy.random.rand(256 * 20 * 10)
    data = data.reshape((256, 20, 10))
    process_data = numpy.random.rand(256 * 20 * 10).reshape((256, 20, 10))
    energy = numpy.linspace(start=3.25, stop=3.69, num=256)

    write_xas(h5_file=h5_file, entry="scan1", sample=sample, energy=energy, mu=data)

    process_norm = PyMca_normalization()
    write_xas_proc(
        h5_file=h5_file,
        entry="scan1",
        process=process_norm,
        results=process_data,
        processing_order=1,
    )
    process_exafs = PyMca_exafs()
    process_data2 = numpy.random.rand(256 * 20 * 10).reshape((256, 20, 10))
    write_xas_proc(
        h5_file=h5_file,
        entry="scan1",
        process=process_exafs,
        results=process_data2,
        processing_order=2,
    )


def get_column_name(dat_file):
    from silx.io.spech5 import SpecH5
    from silx.io.spech5 import SpecFile

    spec_h5 = SpecH5(dat_file)
    spec_file = SpecFile(dat_file)


if __name__ == "__main__":
    get_column_name("specfiledata_tests.dat")



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

"""
This module is used to manage the rsync between files for transfert.
"""

__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "26/06/2019"

from AnyQt.QtWidgets import QApplication
from AnyQt.QtCore import Qt, QUrl, QEvent, pyqtSignal as Signal
from est.core.utils.designpattern import singleton

# TODO: this should be removed


@singleton
class QApplicationManager(QApplication):
    """Return a singleton on the CanvasApplication"""

    fileOpenRequest = Signal(QUrl)

    def __init__(self):
        QApplication.__init__(self, [])
        self.setAttribute(Qt.AA_DontShowIconsInMenus, True)

    def event(self, event):
        if event.type() == QEvent.FileOpen:
            self.fileOpenRequest.emit(event.url())

        return QApplication.event(self, event)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/08/2019"

import unittest
import os
import tempfile
import shutil
from est.core.types import Spectrum, XASObject
from est.core.io import XASWriter
from est.core.types import Dim

try:
    import larch
except ImportError:
    has_larch = False
else:
    has_larch = True
    from est.core.process.larch.autobk import Larch_autobk
    from est.core.process.larch.pre_edge import Larch_pre_edge
    from est.core.process.larch.xftf import Larch_xftf
try:
    import PyMca5
except ImportError:
    has_pymca = False
else:
    has_pymca = True
    from PyMca5.PyMcaDataDir import PYMCA_DATA_DIR
    from est.io.utils.spec import read_spectrum
    from est.core.process.pymca.normalization import PyMca_normalization
    from est.core.process.pymca.exafs import PyMca_exafs
    from est.core.process.pymca.ft import PyMca_ft
    from est.core.process.pymca.k_weight import PyMca_k_weight


@unittest.skipIf(not has_pymca or not has_larch, "pymca and/or larch not installed")
class TestMeldWorkflow(unittest.TestCase):
    """
    Test the following workflow: read_xmu -> pymca normalize ->
    """

    def setUp(self):
        data_file = os.path.join(PYMCA_DATA_DIR, "EXAFS_Cu.dat")
        energy, mu = read_spectrum(
            data_file,
            energy_col_name="Column 1",
            absorption_col_name="Column 2",
        )
        self.spectrum = Spectrum(energy=energy, mu=mu)
        self.xas_obj = XASObject(
            energy=energy, spectra=(self.spectrum,), dim1=1, dim2=1
        )
        assert self.xas_obj.linked_h5_file is not None
        self.output_dir = tempfile.mkdtemp()
        self.h5_file = os.path.join(self.output_dir, "output_file.h5")
        self.configuration_exafs = {
            "Knots": {"Values": (1, 2, 5), "Number": 3, "Orders": [3, 3, 3]},
            "KMin": 0,
            "KMax": 2.3,
        }

    def tearDown(self):
        shutil.rmtree(self.output_dir)

    def test_workflow_normalize_autobk_kweight_ft(self):
        """Test the following workflow:
        pymca normalize -> larch autobk -> kweight -> pymca ft
        """
        self.xas_obj.configuration = {"EXAFS": self.configuration_exafs}
        # pymca normalization
        out = PyMca_normalization()(xas_obj=self.xas_obj)
        # larch autobk
        out = Larch_autobk()(xas_obj=out)
        # k weight
        kweight_process = PyMca_k_weight()
        kweight_process.setConfiguration({"k_weight": 0})
        out = kweight_process(xas_obj=out)
        # pymca ft
        out = PyMca_ft()(xas_obj=out)

        self.assertTrue(out.spectra.data.flat[0].ft.intensity is not None)
        self.assertTrue(len(out.spectra.data.flat[0].ft.intensity) > 1)

    def test_workflow_preedge_exafs_xftf(self):
        """Test the following workflow:
        larch pre_edge -> pymca exafs -> larch xftf
        """
        # larch pre edge
        out = Larch_pre_edge()(self.xas_obj)
        # pymca exafs
        exafs = PyMca_exafs()

        exafs.setConfiguration(self.configuration_exafs)
        out = exafs(xas_obj=out)
        # for now we cannot link xftf because chi is not set by pymca exafs
        spectrum_0 = out.spectra.data.flat[0]
        self.assertFalse(spectrum_0["EXAFSSignal"] is None)
        self.assertFalse(spectrum_0["EXAFSKValues"] is None)
        self.assertFalse(spectrum_0.chi is None)
        self.assertFalse(spectrum_0.k is None)
        # larch xftf
        out = Larch_xftf()(xas_obj=out)

        writer = XASWriter()
        writer.output_file = self.h5_file
        writer(out)


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestMeldWorkflow,):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite



#!/usr/bin/env python
# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""This module describe est applications which are available  through
the silx launcher.

Your environment should provide a command `est`. You can reach help with
`est --help`, and check the version with `est --version`.
"""

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/12/2020"


import logging

logging.basicConfig()

import sys
from silx.utils.launcher import Launcher

try:
    import est._version

    import_version = True
except ImportError:
    import_version = False


def main():
    """Main function of the launcher

    This function is referenced in the setup.py file, to create a
    launcher script generated by setuptools.

    :rtype: int
    :returns: The execution status
    """
    if import_version is False:
        _version = "???"
    else:
        _version = est._version.version
    launcher = Launcher(prog="est", version=_version)
    launcher.add_command(
        "canvas", module_name="est.app.canvas", description="open the orange-canvas"
    )
    launcher.add_command(
        "process",
        module_name="est.app.process",
        description="process a workflow from a .xml.ows file description",
    )
    launcher.add_command(
        "xas-viewer",
        module_name="est.app.xas_viewer",
        description="Open a viewer for browsing xas data",
    )
    launcher.add_command(
        "test", module_name="est.app.test_", description="Launch est unittest"
    )
    launcher.add_command(
        "ows-to-script",
        module_name="est.app.ows_to_script",
        description="Convert a .ows file to python script",
    )

    status = launcher.execute(sys.argv)
    return status


if __name__ == "__main__":
    # executed when using python -m PROJECT_NAME
    status = main()
    sys.exit(status)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017-2019 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/
"""some utils function for executing reprocessing"""


__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "07/18/2019"

import importlib
import logging
from silx.io.utils import h5py_read_dataset

_logger = logging.getLogger(__name__)


def get_process_instance_frm_h5_desc(desc):
    """

    :param dict desc: description of the process to instanciate
    :return: instance of the process to execute, configured from the description
    """
    assert "program" in desc
    assert "class_instance" in desc
    tmp = h5py_read_dataset(desc["class_instance"]).split(".")
    module_name = ".".join(tmp[:-1])
    class_name = tmp[-1]
    try:
        _module = importlib.import_module(module_name)
        instance = getattr(_module, class_name)()
    except Exception as e:
        _logger.warning(
            " ".join(("Fail to instanciate", module_name, class_name, "reason is", e))
        )
        instance = None
    else:
        if "configuration" in desc:
            instance.setConfiguration(desc["configuration"])
    return instance
